/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.model.board;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import sudoku.controller.GameController;
import sudoku.controller.SudokuTransport;
import sudoku.model.games.*;


/**
 * Abstract class with methods for all types of sudoku
 * 
 * @author sk
 */
public abstract class Sudoku implements ISudoku {

    /**
     * Array of all cells of the sudoku.
     */
    protected Cell[][] cells;
    /**
     * A list with all unset cells of the sudoku.
     */
    protected List<Cell> freeCells;
    /**
     * The size of the sudoku: number of cells per row or column.
     * (9 in case of a standard regular sudoku)
     */
    protected int size;
    /**
     * Number of cells in one row/column of a subsquare.
     * (3 in case of a standard regular sudoku)
     */
    protected int root;
    /**
     * List of all areas in which cells are influencing each other.
     * (standard sudoku: 9 rows, 9 columns, 9 subsquares)
     */
    protected List<Area> areas;
    /**
     * The number of different foreground areas this sudoku has.
     */
    protected int foregroundColors = 0;
    /**
     * A random generator used to shuffle the list of free cells.
     */
    private final Random random = new Random();
    /**
     * Name of the sudoku.
     */
    protected String name = "Sudoku";
    /**
     * Type of the sudoku.
     */
    protected GameType type;
    protected boolean isInterrupted = false;
    protected GameController myGC;

    /**
     * The size of the sudoku (number of rows)
     * @return The size of the sudoku
     */
    public int getSize() {
        return size;
    }
    
    /**
     * Number of cells in one row/column of a subsquare
     * (3 in case of a standard regular sudoku)
     * @return Number of cells per subsquare row
     */
    public int getRoot() {
        return root;
    }
    
    /**
     * The name of the sudoku
     * @return The name
     */
    public String getName() {
        return name;
    }
    
    /**
     * Get the type of the Sudoku.
     * @return The type of the sudoku
     */
    public GameType getType() {
        return type;
    }
    
    /**
     * Get the value of a cell.
     * @param x The x coordinate (0-based)
     * @param y The y coordinate (0-based)
     * @return The value of the cell or 0, if unset
     */
    public int getValue(int x, int y) {
        return cells[x][y].getValue();
    }

    /**
     * The list of free cells (cells with no value set).
     * @return The list of free cells
     */
    public List<Cell> getFreeCells() {
        return freeCells;
    }

    /**
     * The constructor.
     * @param size The size of the sudoku (number of rows)
     */
    protected Sudoku(int size, GameController gc) {
        init(size);
        myGC = gc;
    }

    /**
     * Initializes the sudoku.
     * @param size The number of rows
     */
    private void init(int size) {
        areas = new ArrayList<>();
        freeCells = new ArrayList<>();
        this.size = size;
        root = (int) Math.sqrt(size);
        cells = new Cell[size][size];
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                cells[i][j] = new Cell(size, i + 1, j + 1, this);
                freeCells.add(cells[i][j]);
            }
        }
    }

    /**
     * Marks a cell as unset. Adds it to the list of free cells.
     * @param cell The cell to be marked
     */
    @Override
    public void markCellAsFree(Cell cell) {
        if (cell != null) {
            freeCells.add(cell);
        }
    }

    /**
     * Sorts the list of free cells by their count of area membership.
     */
    private void sortList() {
        freeCells.sort((x, y)
                -> {
            if (x.getAreas().size() > y.getAreas().size()) {
                return -1;
            }
            if (x.getAreas().size() < y.getAreas().size()) {
                return +1;
            }
            return 0;
        });
    }

    /**
     * Console representation of the sudoku.
     * @return The string containing the sudoku
     */
    @Override
    public String toString() {
        String res = "";
        String delim = "+-";
        for (int i = 0; i < size; i++) {
            delim += "--";
            if (i % root == root - 1) {
                delim += "+-";
            }
        }
        delim = delim.substring(0, delim.length() - 1) + "\n";

        res = delim;
        for (int i = 0; i < size; i++) {
            res += "| ";
            for (int j = 0; j < size; j++) {
                int v = cells[i][j].getValue();

                res += (char) ((v <= 9) ? v + '0' : v + 'A' - 10) + " ";
                if (j % root == root - 1) {
                    res += "| ";
                }
            }
            res += "\n";
            if (i % root == root - 1) {
                res += delim;
            }

        }
        return res;
    }

    /**
     * Solves the sudoku.
     * @return True if a solution was found, false otherwise
     */
    public void solve() {
        sortList();
        int f = findMinSolutions(1);
        myGC.gotSolution(this, f == 1);
    }

    /**
     * Searches for a given amount of different solutions recursively.
     * Stops when reached the amount.
     * @param solutionCount The amount of solutions we have to find
     * @return 
     */
    public int findMinSolutions(int solutionCount) {
        int sols = 0;
        SudokuState state = getState();
        if (state == SudokuState.DONE) {
            return 1;
        }
        if (state == SudokuState.FAIL) {
            return 0;
        }

        Cell c = Sudoku.this.nextFreeCell();

        for (int i = 1; i <= size; i++) {
            if (!c.setValue(i)) {
                continue;
            }
            sols += findMinSolutions(solutionCount);
            if (sols > 0) {
                if (sols == solutionCount)
                    return sols;
            }
            c.clearValue();
        }
        return sols;
    }
    
    /**
     * Whether this sudoku has exactly one solution.
     * @return True if it has exactly one solution, false otherwise.
     */
    public boolean isDefinite() {
        SudokuState state = this.getState();
        switch(state) {
            case DONE:
                return true;
            case FAIL:
                return false;
        }    
        Sudoku testSudoku = this.copy();
        
        return (testSudoku.findMinSolutions(2) == 2);
    }

    /**
     * Tries to set the cell (x,y) to the given value, if possible.
     * If not possible, the sudoku stays unchanged.
     * @param x Row of the cell (0-based)
     * @param y Column of the cell (0-based)
     * @param value The new value
     * @return True if value could be entered, false otherwise
     */
    public boolean setValue(int x, int y, int value) {
        if (x >= size || y >= size || x < 0 || y < 0) {
            return false;
        }
        if (value < 1 || value > size) {
            return false;
        }
        if (value == 0) {
            cells[x][y].clearValue();
            return true;
        }
        int oldval = cells[x][y].getValue();
        cells[x][y].clearValue();
        if (!cells[x][y].setValue(value)) {
            cells[x][y].clearValue();
            cells[x][y].setValue(oldval);
            return false;
        }
        return true;
    }

    /**
     * The state of this sudoku
     * @return The state
     */
    public SudokuState getState() {
        if (freeCells.size() > 0) {
            return SudokuState.TODO;
        }
        return SudokuState.DONE;
    }

    /**
     * Returns the next free cell in the list of free cells.
     * @return The next free cell, null if there is no more free cell
     */
    public Cell nextFreeCell() {
        return nextFreeCell(false);
    }

    /**
     * Returns the next or a randomly chosen free cell in the list of free cells.
     * @param shuffled Flag if the cell shall be chosen randomly
     * @return A free cell or null (if no free cell available)
     */
    public Cell nextFreeCell(boolean shuffled) {
        if (freeCells.size() > 0) {
            if (shuffled) {
                return freeCells.get(random.nextInt(freeCells.size()));
            } else {
                //SortList();
                return freeCells.get(0);
            }
        }
        return null;
    }
    
    /**
     * Creates a copy of this sudoku.
     * @return The copy of this sudoku
     */
    @Override
    public abstract Sudoku copy();
    
    /**
     * Copies all cell values from this to the given sudoku.
     * @param sudoku
     * @return 
     */
    protected Sudoku copy(Sudoku sudoku) {
        
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                sudoku.cells[i][j].setValue(cells[i][j].getValue());
            }
        }
        return sudoku;
    }
    
    /**
     * Returns a sudoku transport object from this sudoku.
     * @return 
     */
    public SudokuTransport getSudokuTransport() {
        SudokuTransport s = new SudokuTransport();
        s.type = type;
        s.size = size;
        s.foregroundColors = foregroundColors;
        s.values = new int[size][size];
        for (int i = 0; i < size; i++)
            for (int j = 0; j < size; j++)
                s.values[i][j] = cells[i][j].getValue();
        for (Area a : areas)
            s.areas.add(a.getAreaTransport());
        return s;
    }
    
    /**
     * Creates a sudoku from a sudoku transport object.
     * @param st The sudoku transport object
     * @return 
     */
    public static Sudoku getSudoku(SudokuTransport st, GameController gc) {
        Sudoku s = null;
        switch(st.type) {
            case Color: s = new ColorSudoku(st.size, gc); break;
            case Hyper: s = new HyperSudoku(st.size, gc); break;
            case Regular: s = new RegularSudoku(st.size, gc); break;
            case X: s = new XSudoku(st.size, gc); break;
        }
        for (int i = 0; i < st.size; i++)
            for (int j = 0; j < st.size; j++) {
                s.setValue(i, j, st.values[i][j]);
            }
        return s;
    }

    
}
